"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
require("reflect-metadata");
const logger_1 = require("../services/logger");
const server_1 = require("../server/server");
class Injector {
    constructor(server) {
        this.server = server;
        this.temporaryValue = Symbol.for('recursive-injection');
        this.cache = new Map();
        this.cache.set(server_1.Server, () => this.server);
        this.cache.set(logger_1.Logger, () => this.logger);
    }
    get logger() {
        return this.server.logger;
    }
    resolveInjectable(ctorFn) {
        if (!ctorFn) {
            this.logger.error('injector', 'Attempted to inject a falsey type.');
            return;
        }
        if (this.cache.get(ctorFn) === this.temporaryValue) {
            this.logger.error('injector', `Detected circular dependency. Recursive injection of type ${ctorFn.name || ctorFn}`);
            return;
        }
        if (this.cache.has(ctorFn)) {
            return this.cache.get(ctorFn)();
        }
        this.cache.set(ctorFn, this.temporaryValue);
        let inst = this.construct(ctorFn);
        this.cache.set(ctorFn, () => inst);
        return inst;
    }
    construct(ctorFn) {
        let types = Reflect.getOwnMetadata('design:paramtypes', ctorFn);
        if (!types)
            types = [];
        if (types.find(type => type == Object))
            throw new Error(`Could not dependency inject ${ctorFn.name || ctorFn}. Failed to resolve constructor types. Reflected: [${types.map(type => type.name || type).join(', ')}]`);
        let values = types.map(type => this.resolveInjectable(type));
        return new ctorFn(...values);
    }
    provide(provideMeta) {
        let ctorFn = provideMeta.provide;
        if (!ctorFn)
            throw new Error('Attempted to provide a value for a falsey type.');
        if (this.cache.has(ctorFn))
            throw new Error(`Duplicate value provided for ${ctorFn.name || ctorFn}.`);
        let tFn;
        if (this.isClassSource(provideMeta)) {
            this.logger.info('injector', `Providing ${ctorFn.name || ctorFn} using class ${provideMeta.useClass.name || provideMeta.useClass}`);
            let t = this.construct(provideMeta.useClass);
            tFn = () => t;
        }
        else if (this.isValueSource(provideMeta)) {
            this.logger.info('injector', `Providing ${ctorFn.name || ctorFn} using value ${JSON.stringify(provideMeta.useValue)}`);
            let t = provideMeta.useValue;
            tFn = () => t;
        }
        else if (this.isCallbackSource(provideMeta)) {
            this.logger.info('injector', `Providing ${ctorFn.name || ctorFn} using callback ${provideMeta.useCallback.name || provideMeta.useCallback}`);
            tFn = provideMeta.useCallback;
        }
        else
            throw new Error(`Could not resolve dependency injection provider source.`);
        this.cache.set(ctorFn, tFn);
        return this;
    }
    isClassSource(meta) {
        return !!meta.useClass;
    }
    isValueSource(meta) {
        return !!meta.useValue;
    }
    isCallbackSource(meta) {
        return !!meta.useCallback;
    }
}
exports.Injector = Injector;
//# sourceMappingURL=injector.js.map