"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const _ = require("lodash");
const prop_1 = require("../../metadata/orm/prop");
const model_1 = require("../../metadata/orm/model");
const belongs_to_1 = require("../../metadata/orm/associations/belongs-to");
const has_one_1 = require("../../metadata/orm/associations/has-one");
const orm_1 = require("../../decorators/orm");
const transaction_impl_1 = require("./transaction-impl");
class DbImpl {
    constructor(modelFn, model, sequelize, logger) {
        this.modelFn = modelFn;
        this.model = model;
        this.sequelize = sequelize;
        this.logger = logger;
        this.createCopyValsFn();
        this.createTransformQuery();
    }
    transaction(transaction) {
        return __awaiter(this, void 0, void 0, function* () {
            let sqlTransact = transaction && (yield transaction.sync());
            sqlTransact = yield this.sequelize.transaction({ transaction: sqlTransact });
            return new transaction_impl_1.TransactionImpl(sqlTransact);
        });
    }
    create(t, transaction) {
        return __awaiter(this, void 0, void 0, function* () {
            let sqlTransact = transaction && (yield transaction.sync());
            if (t instanceof Array) {
                t = _.cloneDeep(t);
                t.forEach((t, idx, arr) => { arr[idx] = this.transformQueryWhere(t); });
                let results = yield this.model.bulkCreate(t, _.merge({}, { transaction: sqlTransact }));
                return true;
            }
            else {
                t = this.transformQueryWhere(t);
                let result = yield this.model.create(t, _.merge({}, { transaction: sqlTransact }));
                return this.wrapResult(result);
            }
        });
    }
    findById(id, options, transaction) {
        return __awaiter(this, void 0, void 0, function* () {
            let sqlTransact = transaction && (yield transaction.sync());
            if (options)
                options = this.transformQuery(options);
            let result = yield this.model.findById(id, _.merge({}, { transaction: sqlTransact }, options));
            return result && this.wrapResult(result);
        });
    }
    findOne(query, transaction) {
        return __awaiter(this, void 0, void 0, function* () {
            let sqlTransact = transaction && (yield transaction.sync());
            query = this.transformQuery(query);
            let result = yield this.model.findOne(_.merge({}, { transaction: sqlTransact }, query));
            return result && this.wrapResult(result);
        });
    }
    findOrCreate(query, defaults, transaction) {
        return __awaiter(this, void 0, void 0, function* () {
            let sqlTransact = transaction && (yield transaction.sync());
            query = this.transformQueryWhere(query);
            defaults = this.transformQueryWhere(defaults);
            let [result, created] = yield this.model.findOrCreate(_.merge({}, { where: query, defaults: defaults || {}, transaction: sqlTransact }));
            return [result && this.wrapResult(result), created];
        });
    }
    findAndCountAll(query, transaction) {
        return __awaiter(this, void 0, void 0, function* () {
            let sqlTransact = transaction && (yield transaction.sync());
            if (query)
                query = this.transformQuery(query);
            let results = yield this.model.findAndCountAll(_.merge({}, { transaction: sqlTransact }, query));
            return { count: results.count, results: this.wrapResults(results.rows) };
        });
    }
    findAll(query, transaction) {
        return __awaiter(this, void 0, void 0, function* () {
            let sqlTransact = transaction && (yield transaction.sync());
            if (query)
                query = this.transformQuery(query);
            let results = yield this.model.findAll(_.merge({}, { transaction: sqlTransact }, query));
            return this.wrapResults(results);
        });
    }
    all(query, transaction) {
        return this.findAll(query, transaction);
    }
    count(query, transaction) {
        return __awaiter(this, void 0, void 0, function* () {
            let sqlTransact = transaction && (yield transaction.sync());
            if (query)
                query = this.transformQuery(query);
            return yield this.model.count(_.merge({}, { transaction: sqlTransact }, query));
        });
    }
    max(field, transaction) {
        return __awaiter(this, void 0, void 0, function* () {
            let sqlTransact = transaction && (yield transaction.sync());
            return yield this.model.max(field, _.merge({}, { transaction: sqlTransact }));
        });
    }
    min(field, transaction) {
        return __awaiter(this, void 0, void 0, function* () {
            let sqlTransact = transaction && (yield transaction.sync());
            return yield this.model.min(field, _.merge({}, { transaction: sqlTransact }));
        });
    }
    sum(field, transaction) {
        return __awaiter(this, void 0, void 0, function* () {
            let sqlTransact = transaction && (yield transaction.sync());
            return yield this.model.sum(field, _.merge({}, { transaction: sqlTransact }));
        });
    }
    save(t, transaction) {
        return __awaiter(this, void 0, void 0, function* () {
            let [result, created] = yield this.findOrCreate({ id: t.id }, t, transaction);
            return result;
        });
    }
    update(query, replace, returning = false, transaction) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!query) {
                throw new Error(`Db.update query parameter was falsey: ${query}`);
            }
            let sqlTransact = transaction && (yield transaction.sync());
            if (transaction && returning) {
                throw new Error(`Using a transaction to return records when you update them is not yet implemented.`);
            }
            if (this.isId(query))
                query = { where: { id: query } };
            else if (this.isT(query))
                query = { where: { id: query.id } };
            else {
                query = this.transformQuery(query);
                if (returning) {
                    let results = yield this.model.findAll(query);
                    let ids = results.map((result) => result.id);
                    query = { where: { id: { $in: ids } } };
                }
            }
            replace = this.transformQueryWhere(replace);
            let [affected, results] = yield this.model.update(replace, _.merge({}, { transaction: sqlTransact }, query));
            if (returning)
                results = this.wrapResults(yield this.model.findAll(query));
            return [affected, results];
        });
    }
    updateOrCreate(query, defaults, transaction) {
        return __awaiter(this, void 0, void 0, function* () {
            let t = yield this.transaction(transaction);
            let failed = false;
            try {
                let [result, created] = yield this.findOrCreate(query, defaults, t);
                if (!created) {
                    let worked = yield this.update({ where: query }, defaults, false, t);
                    if (!worked)
                        throw new Error("Failed to update or create a model.");
                    let resultOrNull = yield this.findOne(_.merge({}, query, defaults), t);
                    if (!resultOrNull)
                        throw new Error("Updated row, but could not find it afterwards.");
                    result = resultOrNull;
                }
                return [result, created];
            }
            catch (e) {
                failed = true;
                this.logger.error('dbimpl', e);
                yield t.rollback();
                throw e;
            }
            finally {
                if (!failed) {
                    yield t.commit();
                }
            }
        });
    }
    destroy(query, transaction) {
        return __awaiter(this, void 0, void 0, function* () {
            let sqlTransact = transaction && (yield transaction.sync());
            if (this.isId(query))
                query = { where: { id: query } };
            else if (this.isT(query))
                query = { where: { id: query.id } };
            else
                query = this.transformQuery(query);
            return yield this.model.destroy(_.merge({}, { transaction: sqlTransact }, query));
        });
    }
    isId(query) {
        return typeof query === 'number' || typeof query == 'string';
    }
    isT(query) {
        return !!query.id;
    }
    createCopyValsFn() {
        let allProps = [];
        let directTransformFn = (val) => val;
        let props = Reflect.getOwnMetadata(model_1.ModelPropertiesSym, this.modelFn.prototype) || [];
        for (let q = 0; q < props.length; q++) {
            let propName = props[q];
            let propMeta = Reflect.getOwnMetadata(prop_1.PropMetadataSym, this.modelFn.prototype, propName);
            if (!propMeta)
                throw new Error(`Could not find model property metadata for property ${this.modelFn.name || this.modelFn}.${propName}.`);
            let transformFn = directTransformFn;
            if (propMeta.type == orm_1.Types.DATE)
                transformFn = (dateStr) => new Date(dateStr);
            allProps.push({ columnName: propMeta.columnName || propName, propertyName: propName, transformFn: transformFn });
        }
        this.copyVals = function (sql, t) {
            for (let q = 0; q < allProps.length; q++) {
                let propName = allProps[q].propertyName;
                t[propName] = allProps[q].transformFn(sql[propName]);
            }
        };
    }
    static getForeignModelDbImpl(foreignModel) {
        if (!foreignModel || !("db" in foreignModel))
            throw new Error("Cannot get DbImpl from ForeignModel that has not been resolved");
        let staticModel = foreignModel;
        return staticModel.db;
    }
    createTransformQuery() {
        let transforms = this.getTransforms();
        this.createTransformQueryWhere(transforms);
        this.createTransformQueryInclude(transforms);
        this.createTransformResult(transforms);
    }
    transformQuery(query) {
        let result = _.clone(query);
        if (query.where)
            result.where = this.transformQueryWhere(query.where);
        if (query.include)
            result.include = this.transformQueryInclude(query.include);
        return result;
    }
    getTransforms() {
        let transforms = [];
        let belongsTo = Reflect.getOwnMetadata(belongs_to_1.ModelBelongsToAssociationsSym, this.modelFn.prototype) || [];
        for (let q = 0; q < belongsTo.length; q++) {
            let propName = belongsTo[q];
            let propMeta = Reflect.getOwnMetadata(belongs_to_1.BelongsToMetadataSym, this.modelFn.prototype, propName);
            if (!propMeta)
                throw new Error(`Could not find model belongs-to metadata for property ${this.modelFn.name || this.modelFn}.${propName}.`);
            let fkey = propMeta.foreignKey;
            if (fkey && typeof fkey !== 'string')
                fkey = fkey.name;
            if (!fkey)
                throw new Error(`Could not get foreign key for belongs-to property ${this.modelFn.name || this.modelFn}.${propName}`);
            transforms.push({
                type: 'belongs-to',
                columnName: fkey,
                fieldName: propName,
                foreignPkName: 'id',
                foreignDb: () => DbImpl.getForeignModelDbImpl(propMeta.foreignModel)
            });
        }
        let hasOne = Reflect.getOwnMetadata(has_one_1.ModelHasOneAssociationsSym, this.modelFn.prototype) || [];
        for (let q = 0; q < hasOne.length; q++) {
            let propName = hasOne[q];
            let propMeta = Reflect.getOwnMetadata(has_one_1.HasOneMetadataSym, this.modelFn.prototype, propName);
            if (!propMeta)
                throw new Error(`Could not find model has-one metadata for property ${this.modelFn.name || this.modelFn}.${propName}.`);
            let fkey = propMeta.foreignKey;
            if (fkey && typeof fkey !== 'string')
                fkey = fkey.name;
            if (!fkey)
                throw new Error(`Could not get foreign key for has-one property ${this.modelFn.name || this.modelFn}.${propName}`);
            transforms.push({
                type: 'has-one',
                foreignColumnName: fkey,
                fieldName: propName,
                pkName: 'id',
                foreignDb: () => DbImpl.getForeignModelDbImpl(propMeta.foreignModel)
            });
        }
        return transforms;
    }
    createTransformQueryWhere(transforms) {
        this.transformQueryWhere = function (query) {
            if (!query)
                return query;
            query = _.clone(query);
            query['$and'] = this.transformQueryWhere(query['$and']);
            let orVal = query['$or'];
            if (orVal && orVal.length) {
                for (let q = 0; q < orVal.length; q++) {
                    orVal[q] = this.transformQueryWhere(orVal[q]);
                }
            }
            for (let q = 0; q < transforms.length; q++) {
                let transform = transforms[q];
                let fieldVal;
                switch (transform.type) {
                    case 'belongs-to':
                        fieldVal = query[transform.fieldName];
                        if (fieldVal) {
                            if (fieldVal[transform.foreignPkName])
                                fieldVal = fieldVal[transform.foreignPkName];
                            query[transform.columnName] = fieldVal;
                            delete query[transform.fieldName];
                        }
                        break;
                    case 'has-one':
                        fieldVal = query[transform.fieldName];
                        if (fieldVal) {
                            throw new Error(`Not implemented! Cannot include has-one value in where query`);
                        }
                        break;
                    default:
                        throw new Error(`WTF? How did you get here?`);
                }
            }
            return query;
        };
    }
    createTransformQueryInclude(transforms) {
        function getFieldModel(field) {
            for (let i = 0; i < transforms.length; i++) {
                if (field == transforms[i].fieldName)
                    return transforms[i].foreignDb().model;
            }
            return null;
        }
        this.transformQueryInclude = function (fields) {
            if (!fields)
                return fields;
            return fields.map(field => {
                let model = getFieldModel(field);
                if (!model)
                    throw new Error(`Cannot find field ${field} from include query`);
                return { model: model, as: field };
            });
        };
    }
    createTransformResult(transforms) {
        this.transformResult = function (sql, result) {
            result = _.clone(result);
            for (let q = 0; q < transforms.length; q++) {
                let transform = transforms[q];
                let foreignDb = transform.foreignDb();
                switch (transform.type) {
                    case 'belongs-to':
                        if (sql[transform.fieldName]) {
                            let t = new foreignDb.modelFn();
                            foreignDb.copyVals(sql[transform.fieldName], t);
                            result[transform.fieldName] = transform.foreignDb().transformResult(sql[transform.fieldName], t);
                        }
                        else if (sql[transform.columnName]) {
                            result[transform.fieldName] = sql[transform.columnName];
                            delete result[transform.columnName];
                        }
                        break;
                    case 'has-one':
                        if (sql[transform.fieldName]) {
                            let t = new foreignDb.modelFn();
                            foreignDb.copyVals(sql[transform.fieldName], t);
                            result[transform.fieldName] = foreignDb.transformResult(sql[transform.fieldName], t);
                        }
                        break;
                    default:
                        throw new Error(`WTF? How did you get here?`);
                }
            }
            return result;
        };
    }
    fromJson(json) {
        return this.wrapResult(json);
    }
    wrapResult(result) {
        let t = new this.modelFn();
        this.copyVals(result, t);
        return this.transformResult(result, t);
    }
    wrapResults(results) {
        return results.map(result => this.wrapResult(result));
    }
}
exports.DbImpl = DbImpl;
//# sourceMappingURL=db-impl.js.map