"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
require("reflect-metadata");
const Sequelize = require("sequelize");
const model_1 = require("../metadata/orm/model");
const prop_1 = require("../metadata/orm/prop");
const has_many_1 = require("../metadata/orm/associations/has-many");
const belongs_to_1 = require("../metadata/orm/associations/belongs-to");
const has_one_1 = require("../metadata/orm/associations/has-one");
const orm_transform_service_1 = require("../services/orm-transform.service");
const db_impl_1 = require("./impl/db-impl");
const transaction_impl_1 = require("./impl/transaction-impl");
class OrmReflector {
    constructor(server) {
        this.server = server;
        this.models = new Map();
        this.modelsByTableName = new Map();
        let ormTransform = server.injector.resolveInjectable(orm_transform_service_1.OrmTransformService);
        if (!ormTransform)
            throw new Error(`Failed to resolve OrmTransformService. Can't reflect ORM models`);
        this.ormTransform = ormTransform;
    }
    get logger() {
        return this.server.logger;
    }
    init() {
        return __awaiter(this, void 0, void 0, function* () {
            let orm = this.server.meta.orm;
            if (!orm || (typeof orm.enabled !== 'undefined' && !orm.enabled) || !orm.db)
                return;
            let db = orm.db;
            let host = db.host;
            let port = undefined;
            if (typeof host !== 'string') {
                port = host.port;
                host = host.domain;
            }
            let charset = orm.db.charset || 'utf8mb4';
            let charsetCollate = `${charset}_general_ci`;
            this.sql = new Sequelize(db.name, db.user, db.password, {
                host: host,
                dialect: db.dialect || 'mysql',
                dialectOptions: {
                    charset: charset
                },
                pool: db.pool,
                define: {
                    charset: charset,
                    collate: charsetCollate
                },
                logging: (msg, ...extras) => this.logger.info('sql', msg, ...extras),
                port: port
            });
            let models = this.server.meta.models || [];
            this.reflectModels(models);
            this.reflectAssociations(models);
            this.createDbImpls(models);
            if (orm.recreate) {
                if ((process.env.NODE_ENV || '') == 'production')
                    throw new Error('Server launched with config value orm.recreate enabled. As a security feature, this causes a crash when NODE_ENV = production.');
                this.logger.warn('orm', `Warning: recreating database tables. Note: this option should not be enabled in production.`);
            }
            yield this.sync(orm.recreate);
        });
    }
    sync(recreate) {
        return __awaiter(this, void 0, void 0, function* () {
            return yield this.sql.sync({ force: recreate || false });
        });
    }
    transaction(transaction) {
        return __awaiter(this, void 0, void 0, function* () {
            let sqlTransact = transaction && (yield transaction.sync());
            sqlTransact = yield this.sql.transaction({ transaction: sqlTransact });
            return new transaction_impl_1.TransactionImpl(sqlTransact);
        });
    }
    reflectModels(models) {
        for (let q = 0; q < models.length; q++) {
            this.reflectModel(models[q]);
        }
    }
    reflectModel(modelFn) {
        if (this.models.has(modelFn))
            throw new Error(`A model was passed to the orm-reflector twice: ${modelFn.name || modelFn}.`);
        let modelProto = modelFn.prototype;
        let meta = Reflect.getOwnMetadata(model_1.ModelMetadataSym, modelProto);
        if (!meta)
            throw new Error(`Expecting class with @Model decorator, could not reflect model properties for ${modelProto}.`);
        let modelOptions = meta;
        modelOptions = this.ormTransform.transformModel(modelOptions) || modelOptions;
        modelOptions.tableName = modelOptions.tableName || this.ormTransform.transformModelName(modelFn.name) || modelFn.name;
        let dupTable = this.modelsByTableName.get(modelOptions.tableName);
        if (dupTable)
            throw new Error(`Defining multiple models with the same table name! ${dupTable.name || dupTable} and ${modelFn.name || modelFn}`);
        this.modelsByTableName.set(modelOptions.tableName, modelFn);
        let columns = {};
        let props = Reflect.getOwnMetadata(model_1.ModelPropertiesSym, modelProto) || [];
        for (let q = 0; q < props.length; q++) {
            let propName = props[q];
            let propMeta = Reflect.getOwnMetadata(prop_1.PropMetadataSym, modelProto, propName);
            if (!propMeta)
                throw new Error(`Could not find model property metadata for property ${modelFn.name || modelFn}.${propName}.`);
            let columnMeta = propMeta;
            columnMeta = this.ormTransform.transformColumn(columnMeta) || columnMeta;
            columnMeta.field = columnMeta.columnName || this.ormTransform.transformColumnName(propName) || propName;
            columns[propName] = columnMeta;
        }
        let model = this.sql.define(modelOptions.tableName, columns, modelOptions);
        this.models.set(modelFn, model);
    }
    reflectAssociations(models) {
        for (let q = 0; q < models.length; q++) {
            this.reflectModelAssociations(models[q]);
        }
    }
    reflectModelAssociations(modelFn) {
        let model = this.models.get(modelFn);
        if (!model)
            throw new Error(`Could not reflect model associations for a model that failed to be reflected: ${modelFn.name || modelFn}.`);
        let modelProto = modelFn.prototype;
        let meta = Reflect.getOwnMetadata(model_1.ModelMetadataSym, modelProto);
        if (!meta)
            throw new Error(`Expecting class with @Model decorator, could not reflect model properties for ${modelProto}.`);
        let associationTypes = [
            {
                sqlName: 'hasMany',
                msgName: 'has-many',
                associationsSym: has_many_1.ModelHasManyAssociationsSym,
                metadataSym: has_many_1.HasManyMetadataSym,
                transform: (propMeta, propName) => {
                }
            },
            {
                sqlName: 'belongsTo',
                msgName: 'belongs-to',
                associationsSym: belongs_to_1.ModelBelongsToAssociationsSym,
                metadataSym: belongs_to_1.BelongsToMetadataSym,
                transform: (propMeta, propName) => {
                    propMeta.foreignKey = propMeta.foreignKey || this.ormTransform.transformAssociationColumnName(propName) || propName;
                }
            },
            {
                sqlName: 'hasOne',
                msgName: 'has-one',
                associationsSym: has_one_1.ModelHasOneAssociationsSym,
                metadataSym: has_one_1.HasOneMetadataSym,
                transform: (propMeta, propName) => {
                }
            }
        ];
        for (let q = 0; q < associationTypes.length; q++) {
            let def = associationTypes[q];
            let associationNames = Reflect.getOwnMetadata(def.associationsSym, modelProto) || [];
            for (let w = 0; w < associationNames.length; w++) {
                let propName = associationNames[w];
                let propMeta = Reflect.getOwnMetadata(def.metadataSym, modelProto, propName);
                if (!propMeta)
                    throw new Error(`Could not find model ${def.msgName} metadata for property ${modelFn.name || modelFn}.${propName}`);
                let foreignModelFn = this.resolveForeignModelFn(propMeta);
                if (!foreignModelFn)
                    throw new Error(`Could not resolve foreign model for ${def.msgName} association ${modelFn.name || modelFn}.${propName}`);
                let foreignModel = this.models.get(foreignModelFn);
                if (!foreignModel)
                    throw new Error(`Could not create ${def.msgName} association ${modelFn.name || modelFn}.${propName} to model that has not been reflected: ${foreignModelFn.name || foreignModelFn}`);
                let sqlMeta = propMeta;
                sqlMeta = this.ormTransform.transformAssociation(sqlMeta) || sqlMeta;
                if (def.transform)
                    def.transform(sqlMeta, propName);
                model[def.sqlName](foreignModel, propMeta);
            }
        }
    }
    createDbImpls(models) {
        for (let q = 0; q < models.length; q++) {
            let modelFn = models[q];
            let model = this.models.get(modelFn);
            if (!model)
                throw new Error(`Could not reflect model associations for a model that failed to be reflected: ${modelFn.name || modelFn}.`);
            let db = new db_impl_1.DbImpl(modelFn, model, this.sql, this.logger);
            modelFn.db = db;
        }
    }
    isStaticModelT(test) {
        return test && !!test.db;
    }
    isTableNameRef(test) {
        return test.tableName;
    }
    isModelNameRef(test) {
        return test.modelName;
    }
    resolveForeignModelFn(meta) {
        let fmod = meta.foreignModel;
        if (!fmod)
            return undefined;
        if (this.isStaticModelT(fmod))
            return fmod;
        if (typeof fmod === 'function')
            return meta.foreignModel = fmod();
        else if (this.isTableNameRef(fmod))
            return meta.foreignModel = this.modelsByTableName.get(fmod.tableName);
        else if (this.isModelNameRef(fmod)) {
            let modelName = fmod.modelName;
            return meta.foreignModel = [...this.models.keys()].find(model => model.name == modelName);
        }
    }
}
exports.OrmReflector = OrmReflector;
//# sourceMappingURL=reflector.js.map