"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
require("reflect-metadata");
const controller_1 = require("../metadata/router/controller");
const route_1 = require("../metadata/router/route");
const inheritance_hierarchy_1 = require("../util/inheritance-hierarchy");
const join_route_paths_1 = require("../util/join-route-paths");
const wrap_promise_1 = require("../util/wrap-promise");
const http_status_type_1 = require("../util/http-status-type");
require("./extend-request");
class RouterReflector {
    constructor(server, router) {
        this.server = server;
        this.router = router;
        this.controllers = {};
        this.unfinishedRoutes = 0;
        this.requestIndex = 0;
    }
    get logger() {
        return this.server.logger;
    }
    reflectRoutes(controllers) {
        for (let q = 0; q < controllers.length; q++) {
            this.reflectControllerRoutes(controllers[q]);
        }
    }
    reflectControllerRoutes(controllerFn) {
        if (this.controllers[controllerFn])
            throw new Error(`A controller was passed to the router-reflector twice: ${controllerFn}.`);
        let controllerInst = this.controllers[controllerFn] = this.server.injector.resolveInjectable(controllerFn);
        let controllerProto = controllerFn.prototype;
        let meta = Reflect.getOwnMetadata(controller_1.ControllerMetadataSym, controllerProto);
        if (!meta)
            throw new Error(`Expecting class with @Controller decorator, could not reflect routes for ${controllerProto}.`);
        this.logger.info('router', `Reflecting routes for controller ${controllerFn.name}`);
        let routes = this.reflectRouteMeta(controllerProto);
        for (let q = 0; q < routes.length; q++) {
            let [routeFnName, routeMetaArr] = routes[q];
            for (let w = 0; w < routeMetaArr.length; w++) {
                let routeMeta = routeMetaArr[w];
                this.addRoute(controllerInst, routeFnName, meta, routeMeta);
            }
        }
    }
    reflectRouteMeta(controllerProto) {
        let hierarchy = inheritance_hierarchy_1.inhertitanceHierarchy(controllerProto);
        let routeMeta = [];
        for (let r = 0; r < hierarchy.length; r++) {
            let fn = hierarchy[r].prototype;
            let routeNames = Reflect.getOwnMetadata(controller_1.ControllerRoutesSym, fn) || [];
            for (let q = 0; q < routeNames.length; q++) {
                let routeFnName = routeNames[q];
                let routeMetaArr = Reflect.getOwnMetadata(route_1.RouteMetadataSym, fn, routeFnName) || [];
                routeMeta.push([routeFnName, routeMetaArr]);
            }
        }
        return routeMeta;
    }
    addRoute(controller, routeFnName, controllerMeta, routeMeta) {
        let policyDescriptors = [
            ...(this.server.meta.policies || []),
            ...(controllerMeta.policies || []),
            ...(routeMeta.policies || [])
        ];
        let policies = this.resolvePolicies(policyDescriptors);
        let boundRoute = controller[routeFnName].bind(controller);
        let pathPart = routeMeta.path;
        if (controller.transformPathPart)
            pathPart = controller.transformPathPart(pathPart) || pathPart;
        let fullPath = join_route_paths_1.joinRoutePaths(...[
            this.server.meta.path || '',
            controllerMeta.path || '',
            pathPart
        ]);
        if (controller.transformPath)
            fullPath = controller.transformPath(fullPath) || fullPath;
        if (typeof routeMeta.method === 'undefined')
            throw new Error(`Failed to create route ${controller}.${routeFnName}. No method set!`);
        this.logger.verbose('router', `& Adding route ${routeFnName} (${routeMeta.method.toUpperCase()} ${fullPath})`);
        this.router[routeMeta.method](fullPath, this.createFullRouterFn(policies, boundRoute, routeMeta));
    }
    resolvePolicies(descriptors) {
        return descriptors.map((desc) => {
            let key;
            let fn;
            if (this.isPolicyCtor(desc)) {
                key = desc;
                let val = this.server.injector.resolveInjectable(desc);
                if (!val)
                    throw new Error(`Could not resolve dependency for policy: ${desc}`);
                desc = val;
            }
            if (this.isPolicyT(desc)) {
                fn = desc.handle.bind(desc);
            }
            else {
                let handler = desc;
                fn = function (req, res) {
                    return __awaiter(this, void 0, void 0, function* () {
                        yield wrap_promise_1.wrapPromise(handler, req, res);
                    });
                };
            }
            return [key, fn];
        });
    }
    isPolicyCtor(desc) {
        return !this.isPolicyT(desc) && !!desc.prototype.handle;
    }
    isPolicyT(desc) {
        return !!desc.handle;
    }
    createFullRouterFn(policies, boundRoute, meta) {
        const self = this;
        return function (req, res) {
            return __awaiter(this, void 0, void 0, function* () {
                let requestIndex = ++self.requestIndex;
                self.logger.verbose('router', `{${requestIndex}} beginning request: ${req.url}`);
                self.logger.verbose('router', `{${requestIndex}} unfinishedRoutes: ${++self.unfinishedRoutes}`);
                let allResults = [];
                req.policyResults = self.createPolicyResultsFn(policies, allResults);
                let initialStatusCode = res.statusCode;
                for (let q = 0; q < policies.length; q++) {
                    let policy = policies[q];
                    let result;
                    let policyCtor = policy[0];
                    let policyName = (policyCtor && (policyCtor.name || policyCtor)) || '(undefined)';
                    try {
                        self.logger.verbose('router', `{${requestIndex}} awaiting policy ${q + 1}/${policies.length} (${policyName})`);
                        result = yield policy[1](req, res);
                        self.logger.verbose('router', `{${requestIndex}} policy ${policyName} returned with result ${JSON.stringify(result)}`);
                    }
                    catch (e) {
                        self.logger.error('router', `{${requestIndex}} policy (${policyName}) threw an exception. Serving 500 - Internal server error`);
                        self.logger.error('router', e);
                        res.status(http_status_type_1.HTTP_STATUS_INTERNAL_SERVER_ERROR);
                        res.send('Internal server error');
                        self.logger.verbose('router', `{${requestIndex}} ending request. unfinishedRoutes: ${--self.unfinishedRoutes}`);
                        return;
                    }
                    allResults.push(result);
                    if (res.statusCode !== initialStatusCode || res.headersSent)
                        return;
                }
                self.logger.verbose('router', `{${requestIndex}} policies complete, creating transaction`);
                let t = meta.transaction && (yield self.server.transaction());
                let failed = false;
                try {
                    self.logger.verbose('router', `{${requestIndex}} calling route`);
                    yield boundRoute(req, res, t);
                    self.logger.verbose('router', `{${requestIndex}} route complete`);
                }
                catch (e) {
                    self.logger.error('router', `{${requestIndex}} route threw an exception. Serving 500 - Internal server error and rolling back transaction.`);
                    self.logger.error('router', e);
                    res.status(http_status_type_1.HTTP_STATUS_INTERNAL_SERVER_ERROR);
                    res.send('Internal server error');
                    self.logger.verbose('router', `{${requestIndex}} rolling back transaction.`);
                    if (t)
                        yield t.rollback();
                    self.logger.verbose('router', `{${requestIndex}} transaction rolled back. unfinishedRoutes: ${--self.unfinishedRoutes}`);
                    failed = true;
                }
                finally {
                    if (!failed && res.statusCode === initialStatusCode && !res.headersSent) {
                        self.logger.error('router', `{${requestIndex}} route failed to send a response. Serving 404 - Not Found`);
                        res.status(http_status_type_1.HTTP_STATUS_NOT_FOUND);
                        res.send(`Not found.`);
                        self.logger.verbose('router', `{${requestIndex}} ending request. unfinishedRoutes: ${--self.unfinishedRoutes}`);
                    }
                    if (!failed) {
                        self.logger.verbose('router', `{${requestIndex}} committing transaction`);
                        if (t)
                            yield t.commit();
                        self.logger.verbose('router', `{${requestIndex}} transaction committed. ending request. unfinishedRoutes: ${--self.unfinishedRoutes}`);
                    }
                }
            });
        };
    }
    createPolicyResultsFn(policies, allResults) {
        let keys = policies.map(poli => poli[0]);
        return function (policyFn) {
            if (typeof policyFn === 'number')
                return allResults[policyFn];
            for (let q = 0; q < keys.length; q++) {
                if (keys[q] === policyFn)
                    return allResults[q];
            }
            return undefined;
        };
    }
}
exports.RouterReflector = RouterReflector;
//# sourceMappingURL=reflector.js.map