"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const express = require("express");
const bodyParser = require("body-parser");
const injector_1 = require("../core/injector");
const reflector_1 = require("../orm/reflector");
const reflector_2 = require("../services/reflector");
const logger_1 = require("../services/logger");
const reflector_3 = require("../router/reflector");
const wrap_promise_1 = require("../util/wrap-promise");
const http = require("http");
const https = require("https");
const debug_module = require("debug");
let debug = debug_module('express:server');
class Server {
    constructor(_meta) {
        this._meta = _meta;
        this.errorCode = 0;
        this._logger = new logger_1.Logger(this.meta.name || null, this.meta.logLevel);
        this._injector = new injector_1.Injector(this);
        if (_meta.inject) {
            for (let q = 0; q < _meta.inject.length; q++) {
                this._injector.provide(_meta.inject[q]);
            }
        }
    }
    get logger() {
        return this._logger;
    }
    get app() {
        return this._app;
    }
    get injector() {
        return this._injector;
    }
    get meta() {
        return this._meta;
    }
    init() {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                this.logger.info('miter', `Initializing miter server...`);
                if (typeof this.meta.port !== 'undefined')
                    yield this.createExpressApp();
                yield this.reflectOrm();
                yield this.startServices();
                if (typeof this.meta.port !== 'undefined')
                    this.reflectRoutes();
            }
            catch (e) {
                this.logger.error('miter', `FATAL ERROR: Failed to launch server.`);
                this.logger.error('miter', e);
                return;
            }
            if (typeof this.meta.port !== 'undefined')
                this.listen();
        });
    }
    shutdown() {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                try {
                    this.logger.info('miter', `Shutting down miter server...`);
                    if (typeof this.meta.port !== 'undefined')
                        yield this.stopListening();
                    yield this.stopServices();
                }
                finally {
                    this.logger.ShutDown();
                }
            }
            catch (e) {
                this.logger.error('miter', `FATAL ERROR: Failed to gracefully shutdown server.`);
                this.logger.error('miter', e);
            }
        });
    }
    createExpressApp() {
        this._app = express();
        this._app.use(bodyParser.urlencoded({ extended: true }), bodyParser.json());
        if (this.meta.allowCrossOrigin) {
            this.logger.warn('miter', `Warning: server starting with cross-origin policy enabled. This should not be enabled in production.`);
            this._app.use(function (req, res, next) {
                res.header("Access-Control-Allow-Origin", "*");
                res.header("Access-Control-Allow-Headers", req.header("Access-Control-Request-Headers"));
                next();
            });
        }
        if (this.meta.middleware && this.meta.middleware.length) {
            this._app.use(...this.meta.middleware);
        }
    }
    reflectOrm() {
        return __awaiter(this, void 0, void 0, function* () {
            let orm = this.meta.orm;
            if (orm && (typeof orm.enabled === 'undefined' || orm.enabled) && orm.db) {
                this.logger.verbose('orm', `Initializing ORM...`);
                this.ormReflector = new reflector_1.OrmReflector(this);
                yield this.ormReflector.init();
                this.logger.info('orm', `Finished initializing ORM.`);
            }
            else if (this.meta.models && this.meta.models.length) {
                this.logger.warn('orm', `Warning: Models included in server metadata, but no orm configuration defined.`);
            }
        });
    }
    transaction(transaction) {
        return this.ormReflector.transaction(transaction);
    }
    startServices() {
        return __awaiter(this, void 0, void 0, function* () {
            this.logger.verbose('services', `Starting services...`);
            this.serviceReflector = new reflector_2.ServiceReflector(this);
            yield this.serviceReflector.reflectServices(this.meta.services || []);
            this.logger.info('services', `Finished starting services.`);
        });
    }
    stopServices() {
        return __awaiter(this, void 0, void 0, function* () {
            this.logger.verbose('services', `Shutting down services...`);
            yield this.serviceReflector.shutdownServices();
            this.logger.info('services', `Finished shutting down services.`);
        });
    }
    reflectRoutes() {
        this.logger.verbose('router', `Loading routes...`);
        let router = express.Router();
        this.routerReflector = new reflector_3.RouterReflector(this, router);
        this.routerReflector.reflectRoutes(this.meta.controllers || []);
        this.app.use(router);
        this.logger.info('router', `Finished loading routes.`);
    }
    listen() {
        this.logger.info('miter', `Serving`);
        if (this._meta.sslEnabled) {
            this.webServer = https.createServer({ key: this._meta.sslPrivateKey, cert: this._meta.sslCertificate }, this.app);
        }
        else {
            this.webServer = http.createServer(this.app);
        }
        this.webServer.listen(this.meta.port, () => this.onListening());
        this.webServer.on("error", (err) => this.onError(err));
    }
    stopListening() {
        return __awaiter(this, void 0, void 0, function* () {
            this.logger.verbose('miter', `Closing HTTP server...`);
            yield wrap_promise_1.wrapPromise((cb) => {
                if (!this.webServer)
                    return cb();
                this.webServer.close(cb);
            });
            this.logger.info('miter', `Finished closing HTTP server.`);
        });
    }
    onError(error) {
        if (error.syscall !== "listen") {
            throw error;
        }
        let bind = (typeof this.meta.port === "string") ? `pipe ${this.meta.port}` : `port ${this.meta.port}`;
        switch (error.code) {
            case "EACCES":
                this.logger.error('miter', `${bind} requires elevated privileges`);
                this.errorCode = 1;
                this.webServer = undefined;
                this.shutdown();
                break;
            case "EADDRINUSE":
                this.logger.error('miter', `${bind} is already in use`);
                this.errorCode = 1;
                this.webServer = undefined;
                this.shutdown();
                break;
            default:
                throw error;
        }
    }
    onListening() {
        if (!this.webServer)
            throw new Error(`onListening called, but there is no httpServer!`);
        let addr = this.webServer.address();
        let bind = (typeof addr === "string") ? `pipe ${addr}` : `port ${addr.port}`;
        this.logger.info('miter', `Listening on ${bind}`);
    }
}
exports.Server = Server;
//# sourceMappingURL=server.js.map